"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const builtins_1 = __importDefault(require("./builtins"));
const ast_1 = require("../ast/ast");
const object_1 = require("../object/object");
const environment_1 = require("../object/environment");
const lexer_1 = __importDefault(require("../lexer/lexer"));
const parser_1 = __importDefault(require("../parser/parser"));
exports.NULL = new object_1.ONull(), exports.TRUE = new object_1.OBoolean(true), exports.FALSE = new object_1.OBoolean(false);
function Eval(node, env) {
    if (node instanceof ast_1.ASTProgram) {
        return evalProgram(node, env);
    }
    else if (node instanceof ast_1.ImportSpec) {
        return evalImportSpec(node, env);
    }
    else if (node instanceof ast_1.ExpressionStatement) {
        return Eval(node.Expression, env);
    }
    else if (node instanceof ast_1.FloatLiteral) {
        return new object_1.OFloat(node.Value);
    }
    else if (node instanceof ast_1.IntegerLiteral) {
        return new object_1.OInteger(node.Value);
    }
    else if (node instanceof ast_1.StringLiteral) {
        return new object_1.OString(node.Value);
    }
    else if (node instanceof ast_1.RangeLiteral) {
        let left = Eval(node.Left, env);
        if (isError(left))
            return left;
        let right = Eval(node.Right, env);
        if (isError(right))
            return right;
        return evalInfixExpression(node.Operator, left, right);
    }
    else if (node instanceof ast_1.Comment) {
        return new object_1.OComment(node.Value);
    }
    else if (node instanceof ast_1.AstBoolean) {
        return nativeBoolToBooleanObject(node.Value);
    }
    else if (node instanceof ast_1.PrefixExpression) {
        let right = Eval(node.Right, env);
        if (isError(right))
            return right;
        return evalPrefixExpression(node.Operator, right);
    }
    else if (node instanceof ast_1.InfixExpression) {
        let left = Eval(node.Left, env);
        if (isError(left))
            return left;
        let right = Eval(node.Right, env);
        if (isError(right))
            return right;
        return evalInfixExpression(node.Operator, left, right);
    }
    else if (node instanceof ast_1.BlockStatement) {
        return evalBlockStatement(node, env);
    }
    else if (node instanceof ast_1.IfExpression) {
        return evalIfExpression(node, env);
    }
    else if (node instanceof ast_1.ReturnStatement) {
        let val = Eval(node.ReturnValue, env);
        if (isError(val))
            return val;
        return val ? new object_1.ReturnValue(val) : null;
    }
    else if (node instanceof ast_1.LetStatement) {
        if (node.Index)
            return evalLetStatement(node, env);
        let val = Eval(node.Value, env);
        if (isNotError(val))
            env.Set(node.Name.Value, val);
        return val;
    }
    else if (node instanceof ast_1.Identifier) {
        return evalIdentifier(node, env);
    }
    else if (node instanceof ast_1.IncrementExpression) {
        return evalIncrementIdentifier(node, env);
    }
    else if (node instanceof ast_1.DecrementExpression) {
        return evalDecrementIdentifier(node, env);
    }
    else if (node instanceof ast_1.WhileLiteral) {
        return evalWhile(node.Expression, node.Body, env);
    }
    else if (node instanceof ast_1.ForLiteral) {
        return evalFor(node.Initiate, node.Check, node.Iterate, node.Body, env);
    }
    else if (node instanceof ast_1.FunctionLiteral) {
        return new object_1.OFunction(node.Parameters, node.Body, env);
    }
    else if (node instanceof ast_1.CallExpression) {
        let func = Eval(node.Function, env);
        if (isNotError(func)) {
            let args = evalExpressions(node.Arguments, env);
            if (args.length === 1 && isError(args[0])) {
                return args[0];
            }
            return applyFunction(env, func, args);
        }
        return func;
    }
    else if (node instanceof ast_1.ArrayLiteral) {
        let elements = evalExpressions(node.Elements, env);
        if (elements.length === 1 && isError(elements[0])) {
            return elements[0];
        }
        return new object_1.OArray(elements);
    }
    else if (node instanceof ast_1.IndexExpression) {
        let left = Eval(node.Left, env);
        if (isError(left))
            return left;
        let index = Eval(node.Index, env);
        let rightIndex = Eval(node.RightIndex, env);
        return evalIndexExpression(left, index, node.HasColon, rightIndex);
    }
    else if (node instanceof ast_1.HashLiteral) {
        return evalHashLiteral(node, env);
    }
    return null;
}
exports.default = Eval;
function evalProgram(program, env) {
    let result = null;
    for (let statement of program.Statements) {
        result = Eval(statement, env);
        if (result instanceof object_1.ReturnValue) {
            return result.Value;
        }
        else if (result instanceof object_1.OError) {
            return result;
        }
    }
    return result;
}
function evalImportSpec(importSpec, env) {
    let globalEnv = env.GlobalEnv();
    if (!globalEnv.FileLoader)
        return exports.NULL;
    let path = importSpec.Path.Value;
    let content = '';
    let cacheKey = '_files.' + path;
    let cache = globalEnv.Get(cacheKey);
    if (cache) {
        content = cache.toString();
    }
    else {
        content = globalEnv.FileLoader.Load(path);
        globalEnv.Set(cacheKey, new object_1.OString(content));
    }
    let l = new lexer_1.default(content);
    let p = new parser_1.default(l);
    let program = p.ParseProgram();
    if (importSpec.Name) {
        let evaluated = Eval(program, environment_1.NewEnclosedEnvironment(env));
        env.Set(importSpec.Name.Value, evaluated);
    }
    else {
        Eval(program, env);
    }
    return null;
}
function evalBlockStatement(program, env) {
    let result = null;
    for (let statement of program.Statements) {
        result = Eval(statement, env);
        if (result !== null) {
            let rt = result.Type();
            if (rt === object_1.RETURN_VALUE_OBJ || rt === object_1.ERROR_OBJ) {
                return result;
            }
        }
    }
    return result;
}
function evalLetStatement(stmt, env) {
    if (!stmt.Index)
        return exports.NULL;
    let name = stmt.Name;
    let hash = evalIdentifier(stmt.Name, env);
    let key = Eval(stmt.Index.Index, env);
    if (!key)
        return exports.NULL;
    let val = Eval(stmt.Value, env);
    let pair = new object_1.HashPair(key, val ? val : exports.NULL);
    hash.Pairs.set(key.HashKey().Match, pair);
    env.Set(name.Value, hash);
    return val;
}
function evalWhile(expression, body, env) {
    let result = null;
    let loopCount = 0;
    while (true) {
        result = Eval(expression, env);
        if (++loopCount > 1000000) {
            return newError('loop count of 1,000,000 exeeded');
        }
        else if (result instanceof object_1.ONull) {
            break;
        }
        else if (result instanceof object_1.OBoolean && result.Value === true) {
            Eval(body, env);
        }
        else if (result instanceof object_1.OBoolean && result.Value === false) {
            break;
        }
        else {
            break;
        }
    }
    return result;
}
function evalFor(initiate, check, iterate, body, env) {
    let loopCount = 0;
    Eval(initiate, env);
    while (true) {
        let result = Eval(check, env);
        if (++loopCount > 1000000) {
            return newError('loop count of 1,000,000 exeeded');
        }
        else if (result instanceof object_1.ONull) {
            break;
        }
        else if (result instanceof object_1.OBoolean && result.Value === true) {
            Eval(body, env);
        }
        else if (result instanceof object_1.OBoolean && result.Value === false) {
            break;
        }
        else {
            break;
        }
        Eval(iterate, env);
    }
    return exports.NULL;
}
function nativeBoolToBooleanObject(input) {
    if (input) {
        return exports.TRUE;
    }
    return exports.FALSE;
}
function evalPrefixExpression(operator, right) {
    switch (operator) {
        case '!':
            return evalBangOperatorExpression(right);
        case '-':
            return evalMinusPrefixOperatorExpression(right);
        case '~':
            return evalNotPrefixOperatorExpression(right);
        default:
            return newError('unknown operator: %s%s', operator, right ? right.Type() : null);
    }
}
function evalBangOperatorExpression(right) {
    if (!right)
        return exports.FALSE;
    if (right instanceof object_1.ONull) {
        return exports.TRUE;
    }
    else if (right instanceof object_1.OBoolean && right.Value === true) {
        return exports.FALSE;
    }
    else if (right instanceof object_1.OBoolean && right.Value === false) {
        return exports.TRUE;
    }
    return exports.FALSE;
}
function evalMinusPrefixOperatorExpression(right) {
    if (!right)
        return exports.FALSE;
    if (right.Type() !== object_1.INTEGER_OBJ && right.Type() !== object_1.FLOAT_OBJ) {
        return newError('unknown operator: -%s', right.Type());
    }
    else if (!(right instanceof object_1.OInteger) && !(right instanceof object_1.OFloat)) {
        return exports.NULL;
    }
    let value = right.Value;
    if (right instanceof object_1.OFloat)
        return new object_1.OFloat(-value);
    return new object_1.OInteger(-value);
}
function evalNotPrefixOperatorExpression(right) {
    if (!right)
        return exports.FALSE;
    if (right.Type() !== object_1.INTEGER_OBJ) {
        return newError('unknown operator: -%s', right.Type());
    }
    else if (!(right instanceof object_1.OInteger)) {
        return exports.NULL;
    }
    let value = right.Value;
    return new object_1.OInteger(~value);
}
function evalInfixExpression(operator, left, right) {
    if (left instanceof object_1.OInteger && right instanceof object_1.OInteger) {
        return evalIntegerInfixExpression(operator, left, right);
    }
    else if (left instanceof object_1.OFloat && right instanceof object_1.OFloat) {
        return evalFloatInfixExpression(operator, left, right);
    }
    else if (left instanceof object_1.OInteger && right instanceof object_1.OFloat) {
        return evalFloatInfixExpression(operator, left, right);
    }
    else if (left instanceof object_1.OFloat && right instanceof object_1.OInteger) {
        return evalFloatInfixExpression(operator, left, right);
    }
    else if (left instanceof object_1.OString && right instanceof object_1.OString) {
        return evalStringInfixExpression(operator, left, right);
    }
    else if (operator === '==') {
        return nativeBoolToBooleanObject(left === right);
    }
    else if (operator === '!=') {
        return nativeBoolToBooleanObject(left !== right);
    }
    else if (operator === 'and' || operator === '&&') {
        if (!left || !(left instanceof object_1.OBoolean))
            return nativeBoolToBooleanObject(false);
        if (!right || !(right instanceof object_1.OBoolean))
            return nativeBoolToBooleanObject(false);
        return nativeBoolToBooleanObject(left.Value && right.Value);
    }
    else if (operator === 'or' || operator === '||') {
        if (left && left instanceof object_1.OBoolean && left.Value)
            return nativeBoolToBooleanObject(true);
        if (right && right instanceof object_1.OBoolean && right.Value)
            return nativeBoolToBooleanObject(true);
        return nativeBoolToBooleanObject(false);
    }
    else if (left !== null && right !== null && left.Type() !== right.Type()) {
        return newError('type mismatch: %s %s %s', left.Type(), operator, right.Type());
    }
    else {
        return newError('unknown operator: %s %s %s', left ? left.Type() : null, operator, right ? right.Type() : null);
    }
}
function evalIntegerInfixExpression(operator, left, right) {
    let leftVal = left.Value;
    let rightVal = right.Value;
    switch (operator) {
        case '+':
            return new object_1.OInteger(leftVal + rightVal);
        case '-':
            return new object_1.OInteger(leftVal - rightVal);
        case '*':
            return new object_1.OInteger(leftVal * rightVal);
        case '/':
            return new object_1.OInteger(leftVal / rightVal);
        case '%':
            return new object_1.OInteger(leftVal % rightVal);
        case '&':
            return new object_1.OInteger(leftVal & rightVal);
        case '|':
            return new object_1.OInteger(leftVal | rightVal);
        case '^':
            return new object_1.OInteger(leftVal ^ rightVal);
        case '<<':
            return new object_1.OInteger(leftVal << rightVal);
        case '>>':
            return new object_1.OInteger(leftVal >> rightVal);
        case '>>>':
            return new object_1.OInteger(leftVal >>> rightVal);
        case '**':
            return new object_1.OInteger(leftVal ** rightVal);
        case '<':
            return nativeBoolToBooleanObject(leftVal < rightVal);
        case '>':
            return nativeBoolToBooleanObject(leftVal > rightVal);
        case '<=':
            return nativeBoolToBooleanObject(leftVal <= rightVal);
        case '>=':
            return nativeBoolToBooleanObject(leftVal >= rightVal);
        case '==':
            return nativeBoolToBooleanObject(leftVal === rightVal);
        case '!=':
            return nativeBoolToBooleanObject(leftVal !== rightVal);
        case '..':
            return evalRangeObject(leftVal, rightVal, false);
        case '...':
            return evalRangeObject(leftVal, rightVal, true);
        default:
            return newError('unknown operator: %s %s %s', left.Type(), operator, right.Type());
    }
}
function evalRangeObject(left, right, include) {
    let arr = [];
    right = include ? right + 1 : right;
    for (let i = left; i < right; i++) {
        arr.push(new object_1.OInteger(i));
    }
    return new object_1.OArray(arr);
}
function evalFloatInfixExpression(operator, left, right) {
    let leftVal = left.Value;
    let rightVal = right.Value;
    switch (operator) {
        case '+':
            return new object_1.OFloat(leftVal + rightVal);
        case '-':
            return new object_1.OFloat(leftVal - rightVal);
        case '*':
            return new object_1.OFloat(leftVal * rightVal);
        case '/':
            return new object_1.OFloat(leftVal / rightVal);
        case '%':
            return new object_1.OFloat(leftVal % rightVal);
        case '**':
            return new object_1.OFloat(leftVal ** rightVal);
        case '<':
            return nativeBoolToBooleanObject(leftVal < rightVal);
        case '>':
            return nativeBoolToBooleanObject(leftVal > rightVal);
        case '<=':
            return nativeBoolToBooleanObject(leftVal <= rightVal);
        case '>=':
            return nativeBoolToBooleanObject(leftVal >= rightVal);
        case '==':
            return nativeBoolToBooleanObject(leftVal === rightVal);
        case '!=':
            return nativeBoolToBooleanObject(leftVal !== rightVal);
        default:
            return newError('unknown operator: %s %s %s', left.Type(), operator, right.Type());
    }
}
function evalStringInfixExpression(operator, left, right) {
    if (operator === '+') {
        let leftVal = left.Value;
        let rightVal = right.Value;
        return new object_1.OString(leftVal + rightVal);
    }
    else if (operator === '==') {
        let leftVal = left.Value;
        let rightVal = right.Value;
        return new object_1.OBoolean(leftVal === rightVal);
    }
    else if (operator === '<') {
        let leftVal = left.Value;
        let rightVal = right.Value;
        return new object_1.OBoolean(leftVal < rightVal);
    }
    else if (operator === '>') {
        let leftVal = left.Value;
        let rightVal = right.Value;
        return new object_1.OBoolean(leftVal > rightVal);
    }
    return newError('unknown operator: %s %s %s', left.Type(), operator, right.Type());
}
function evalIfExpression(ie, env) {
    let condition = Eval(ie.Condition, env);
    if (isError(condition))
        return condition;
    if (isTruthy(condition)) {
        return Eval(ie.Consequence, env);
    }
    else if (ie.Alternative !== null) {
        return Eval(ie.Alternative, env);
    }
    else {
        return exports.NULL;
    }
}
function evalIdentifier(node, env) {
    let val = env.Get(node.Value);
    if (val) {
        return val;
    }
    if (builtins_1.default[node.Value]) {
        return builtins_1.default[node.Value];
    }
    return newError('identifier not found: %s at %s', node.Value, node.Token.Position.String());
}
function evalIncrementIdentifier(node, env) {
    let val = env.Get(node.Name.Value);
    if (val) {
        let incremented = new object_1.OInteger(val.toValue() + 1);
        env.Set(node.Name.Value, incremented);
        return node.Prefix ? incremented : val;
    }
    return newError('identifier not found: %s at %s', node.Name.Value, node.Token.Position.String());
}
function evalDecrementIdentifier(node, env) {
    let val = env.Get(node.Name.Value);
    if (val) {
        let decremented = new object_1.OInteger(val.toValue() - 1);
        env.Set(node.Name.Value, decremented);
        return node.Prefix ? decremented : val;
    }
    return newError('identifier not found: %s at %s', node.Name.Value, node.Token.Position.String());
}
function evalExpressions(exps, env) {
    let result = [];
    for (let e of exps) {
        let evaluated = Eval(e, env);
        if (isNotError(evaluated)) {
            result.push(evaluated);
            continue;
        }
        if (isError(evaluated)) {
            return [evaluated];
        }
    }
    return result;
}
function evalIndexExpression(left, index, hasColon, rightIndex = null) {
    if (left === null)
        return exports.NULL;
    if (left.Type() === object_1.ARRAY_OBJ) {
        return evalArrayIndexExpression(left, index, hasColon, rightIndex);
    }
    else if (left.Type() === object_1.STRING_OBJ) {
        return evalStringIndexExpression(left, index, hasColon, rightIndex);
    }
    else if (index !== null && left.Type() === object_1.HASH_OBJ) {
        return evalHashIndexExpression(left, index);
    }
    else {
        return newError('index operator not supported: %s[%s:%s]', left.Type(), index ? index.Type() : null, rightIndex ? rightIndex.Type() : null);
    }
}
function evalStringIndexExpression(str, index, hasColon = false, rightIndex) {
    let strObject = str;
    let idx = index ? index.Value : null;
    let rIdx = rightIndex ? rightIndex.Value : null;
    let newStr = strObject.Inspect();
    if (idx && hasColon && !rIdx) {
        return new object_1.OString(newStr.substring(idx));
    }
    else if (idx && !rIdx) {
        return new object_1.OString(newStr[idx]);
    }
    else if (idx && rIdx) {
        return new object_1.OString(newStr.substring(idx, rIdx));
    }
    else if (rIdx && hasColon && !idx) {
        return new object_1.OString(newStr.substring(0, rIdx));
    }
    return exports.NULL;
}
function evalArrayIndexExpression(array, index, hasColon = false, rightIndex) {
    let arrayObject = array;
    let idx = index ? index.Value : null;
    let rIdx = rightIndex ? rightIndex.Value : null;
    if (idx && hasColon && !rIdx) {
        return new object_1.OArray(arrayObject.Elements.slice(idx));
    }
    else if (idx && !rIdx) {
        return arrayObject.Elements[idx];
    }
    else if (idx && rIdx) {
        return new object_1.OArray(arrayObject.Elements.slice(idx, rIdx));
    }
    else if (rIdx && hasColon && !idx) {
        return new object_1.OArray(arrayObject.Elements.slice(0, rIdx));
    }
    return exports.NULL;
}
function evalHashLiteral(node, env) {
    let pairs = new Map();
    node.Pairs.forEach((valueNode, keyNode) => {
        let key = Eval(keyNode, env);
        if (isError(key) || !key) {
            return key;
        }
        let hashKey = key;
        if (!(hashKey instanceof object_1.OBoolean || hashKey instanceof object_1.OInteger || hashKey instanceof object_1.OString)) {
            return newError('unusable as hash key: %s', hashKey.Type());
        }
        let value = Eval(valueNode, env);
        if (isError(value) || !value) {
            return value;
        }
        let hashed = hashKey.HashKey();
        pairs.set(hashed.Match, new object_1.HashPair(hashKey, value));
    });
    return new object_1.OHash(pairs);
}
function evalHashIndexExpression(hash, index) {
    let hashObject = hash;
    if (!(hashObject instanceof object_1.OHash)) {
        return newError('unusable as hash object: %s', hashObject.Type());
    }
    let key = index;
    if (!(key instanceof object_1.OBoolean || key instanceof object_1.OInteger || key instanceof object_1.OString)) {
        return newError('unusable as hash key: %s', key.Type());
    }
    let pair = hashObject.Pairs.get(key.HashKey().Match);
    if (!pair) {
        return exports.NULL;
    }
    return pair.Value;
}
function applyFunction(env, fn, args) {
    let func = fn;
    if (func instanceof object_1.OFunction) {
        let extendedEnv = extendedFunctionEnv(func, args);
        let evaluated = Eval(func.Body, extendedEnv);
        return unwrapReturnValue(evaluated);
    }
    else if (func instanceof object_1.Builtin) {
        return func.Fn(env, ...args);
    }
    return newError('not a function: %s', fn.Type());
}
exports.applyFunction = applyFunction;
function extendedFunctionEnv(fn, args) {
    let env = environment_1.NewEnclosedEnvironment(fn.Env);
    for (let [paramIdx, param] of fn.Parameters.entries()) {
        env.Set(param.Value, args[paramIdx]);
    }
    return env;
}
function unwrapReturnValue(obj) {
    let returnValue = obj;
    if (returnValue instanceof object_1.ReturnValue) {
        return returnValue.Value;
    }
    return obj;
}
function isTruthy(obj) {
    switch (obj) {
        case exports.NULL:
            return false;
        case exports.TRUE:
            return true;
        case exports.FALSE:
            return false;
        default:
            return true;
    }
}
function newError(format, ...args) {
    return new object_1.OError(printf(format, ...args));
}
exports.newError = newError;
function printf(...args) {
    return [...args].reduce((p, c) => p.replace(/%s/, c));
}
function isError(obj) {
    if (obj !== null) {
        return obj.Type() === object_1.ERROR_OBJ;
    }
    return false;
}
function isNotError(obj) {
    if (obj !== null) {
        return obj.Type() !== object_1.ERROR_OBJ;
    }
    return true;
}
//# sourceMappingURL=evaluator.js.map