"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const process_1 = __importDefault(require("process"));
const readline_1 = __importStar(require("readline"));
const lexer_1 = __importDefault(require("../lexer/lexer"));
const parser_1 = __importDefault(require("../parser/parser"));
const evaluator_1 = __importDefault(require("../evaluator/evaluator"));
const environment_1 = require("../object/environment");
const object_1 = require("../object/object");
const PROMPT = '>> ';
let inputHistory = [];
let inputText;
function Start(input, output, errors, quiet = false) {
    let env = environment_1.NewNodeEnvironment();
    env.Logger.Follow((date, messages) => {
        console.log('\n', date, ...messages);
    });
    let lastOutput;
    let posX = 0;
    if (input && input.setRawMode) {
        readline_1.default.emitKeypressEvents(input);
        input.setRawMode(true);
    }
    var rl = readline_1.default.createInterface({ input, output, prompt: PROMPT, terminal: false });
    if (!quiet)
        rl.prompt();
    let inputHistoryIndex = 0;
    input.on('keypress', function (s, key) {
        if (key.sequence === '\u0003') {
            process_1.default.exit(0);
        }
        if (key.sequence === '\u0004') {
            process_1.default.exit(0);
        }
        if (key.sequence === '\u0001') {
            readline_1.cursorTo(output, PROMPT.length);
            posX = 0;
        }
        else if (key.sequence === '\u0005') {
            readline_1.cursorTo(output, PROMPT.length + inputText.length);
            posX = inputText.length;
        }
        else if (key.sequence === '\u0015') {
            inputText = inputText.substring(posX, inputText.length);
            posX = 0;
            readline_1.clearLine(output, 0);
            readline_1.cursorTo(output, 0);
            rl.prompt();
            output.write(inputText);
            readline_1.cursorTo(output, posX + PROMPT.length);
        }
        else if (key.sequence === '\u000b') {
            inputText = inputText.substring(0, posX - 1) + inputText.substring(posX, inputText.length);
            readline_1.clearLine(output, 0);
            readline_1.cursorTo(output, 0);
            rl.prompt();
            output.write(inputText);
            readline_1.cursorTo(output, posX + PROMPT.length - 1);
            posX = inputText.length;
        }
        else if (key.sequence === '\u001a') {
            readline_1.clearLine(output, 0);
            readline_1.cursorTo(output, 0);
            rl.prompt();
            posX = 0;
            inputText = '';
        }
        else if (key.name === 'left') {
            if (posX > 0) {
                readline_1.moveCursor(output, -1, 0);
                posX--;
            }
        }
        else if (key.name === 'right') {
            if (posX < inputText.length) {
                readline_1.moveCursor(output, +1, 0);
                posX++;
            }
        }
        else if (key.name === 'up') {
            if (inputHistory.length > 0) {
                if (inputHistoryIndex < inputHistory.length) {
                    inputHistoryIndex++;
                }
                else {
                    return;
                }
                let inputItem = inputHistory[inputHistory.length - inputHistoryIndex];
                inputText = inputItem;
                readline_1.clearLine(output, 0);
                readline_1.cursorTo(output, 0);
                rl.prompt();
                rl.write(inputText);
                output.write(inputText);
                posX = inputText.length;
            }
        }
        else if (key.name === 'down') {
            if (inputHistory.length > 0) {
                let inputItem;
                if (inputHistoryIndex > 1) {
                    inputHistoryIndex--;
                    inputItem = inputHistory[inputHistory.length - inputHistoryIndex];
                }
                else if (inputHistoryIndex > 0) {
                    inputHistoryIndex--;
                    inputItem = '';
                }
                else {
                    return;
                }
                inputText = inputItem;
                readline_1.clearLine(output, 0);
                readline_1.cursorTo(output, 0);
                rl.prompt();
                rl.write(inputText);
                output.write(inputText);
                posX = inputText.length;
            }
        }
        else if (key.name === 'backspace') {
            if (inputText) {
                inputText = inputText.substring(0, posX - 1) + inputText.substring(posX, inputText.length);
            }
            else {
                inputText = '';
            }
            readline_1.clearLine(output, 0);
            readline_1.cursorTo(output, 0);
            rl.prompt();
            output.write(inputText);
            readline_1.cursorTo(output, posX + PROMPT.length - 1);
            if (posX > 1)
                posX--;
        }
        else if (key.name === 'return' || key.name === 'enter') {
            inputHistoryIndex = 0;
            output.write('\n');
            rl.prompt();
            inputText = '';
            posX = inputText.length;
        }
        else if (key.ctrl === false && key.meta === false && key.shift === false) {
            if (inputText !== undefined && inputText !== '') {
                inputText = inputText.substring(0, posX) + key.sequence + inputText.substring(posX, inputText.length);
            }
            else {
                inputText = key.sequence;
            }
            posX++;
            readline_1.clearLine(output, 0);
            readline_1.cursorTo(output, 0);
            rl.prompt();
            output.write(inputText);
            readline_1.cursorTo(output, posX + PROMPT.length);
        }
        else {
            output.write(key.sequence);
            inputText += key.sequence;
            posX = inputText.length;
        }
    });
    rl.on('line', function (line) {
        line = inputText !== undefined ? inputText : line;
        switch (line) {
            case 'exit':
            case 'quit':
            case 'q':
                rl.close();
                process_1.default.exit(0);
                break;
            default:
                if (line !== '')
                    inputHistory.push(line);
                let l = new lexer_1.default(line);
                let p = new parser_1.default(l);
                let program = p.ParseProgram();
                if (p.Errors().length !== 0) {
                    printParserErrors(errors, p.Errors());
                    return;
                }
                let evaluated = evaluator_1.default(program, env);
                if (evaluated !== null) {
                    lastOutput = evaluated;
                    if (!quiet)
                        output.write('\n' + evaluated.Inspect());
                }
                break;
        }
    });
    rl.on('end', () => {
        if (lastOutput instanceof object_1.OInteger)
            process_1.default.exit(lastOutput.Value);
        rl.close();
        process_1.default.exit(0);
    });
    rl.on('close', () => {
        rl.close();
        process_1.default.exit(0);
    });
    rl.on('error', (err) => {
        errors.write(`\x1b[31m==> \t${err}\n\x1b[0m`);
    });
}
exports.default = Start;
function printParserErrors(output, errs) {
    output.write(`\n\x1b[31m`);
    output.write('Woops! We ran into some monkey business here!\n');
    output.write('    Parser errors:\n');
    errs.forEach((msg) => output.write(`\t${msg}\n`));
    output.write('\x1b[0m');
}
//# sourceMappingURL=repl.js.map